import collections as c
import re

import git_graph.git_functions as gf


def build_git_trees(path, trees):
    result = c.defaultdict(list)
    pattern = "(tree|blob) (.+)\t(.+)"
    for each_tree in trees:
        for each_line in gf.read_git_file(path, each_tree):
            match = re.search(pattern, each_line)
            if match:
                result[each_tree].append((match.group(2), match.group(3)))
    return result


def build_git_commits(path, commits):
    result = c.defaultdict(list)
    pattern = "(tree|parent) (.+)"
    for each_commit in commits:
        for each_line in gf.read_git_file(path, each_commit):
            match = re.search(pattern, each_line)
            if match:
                result[each_commit].append(match.group(2))
    return result


def build_git_remote_servers(remote_branches, remote_heads):
    result = c.defaultdict(list)
    for rb in remote_branches:
        server = rb[: rb.find("/")]
        result[server].append(rb)
    for rb in remote_heads:
        server = rb[: rb.find("/")]
        result[server].append(rb)
    return result


def build_git_annotated_tags(path, annotated_tags):
    result = {}
    pattern = "(object) (.+)"
    for each_annotated_tag in annotated_tags:
        for each_line in gf.read_git_file(path, each_annotated_tag):
            match = re.search(pattern, each_line)
            if match:
                result[each_annotated_tag] = match.group(2)
    return result


class GitGraph:
    def __init__(self, path):
        self.path = path
        # b: -> nothing
        self.blobs = []
        # t: -> 0 to N blobs and 0 to N trees
        self.trees = c.defaultdict(list)
        # c: -> 1 tree and 0 to N commits
        self.commits = c.defaultdict(list)
        # l: -> 1 commit
        self.local_branches = {}
        # h: -> 1 local_branch
        self.local_head = ("", "")
        # r: -> 1 commit
        self.remote_branches = {}
        # d: -> 1 remote branch
        self.remote_heads = {}
        # s: -> 1 to N remote_branches
        self.remote_servers = c.defaultdict(list)
        # a: -> 1 commit
        self.annotated_tags = {}
        # g: -> 1 commit or 1 annotated_tag
        self.tags = {}
        # u: -> 1 local_branch
        self.upstreams = {}

    def build_graph(self):
        blobs, trees, commits, annotated_tags = gf.get_git_objects(self.path)
        local_branches, remote_branches, tags = gf.get_git_references(self.path)
        local_head = gf.get_git_local_head(self.path)
        remote_heads = gf.get_git_remote_heads(self.path)
        upstreams = gf.get_git_upstreams(self.path)

        self.blobs = blobs
        self.trees = build_git_trees(self.path, trees)
        self.commits = build_git_commits(self.path, commits)
        self.local_branches = local_branches
        self.local_head = ("HEAD", local_head)
        self.remote_branches = remote_branches
        self.remote_heads = remote_heads
        self.remote_servers = build_git_remote_servers(remote_branches, remote_heads)
        self.annotated_tags = build_git_annotated_tags(self.path, annotated_tags)
        self.tags = tags
        self.upstreams = upstreams
        return self





class BaseNode:
    def __init__(self, data):
        self.data = data

    def __repr__(self):
        return repr(self.data)

    def __str__(self):
        return str(self.data)

    def __eq__(self, other):
        if not isinstance(other, BaseNode):
            return False
        return self.data == other.data
    def __hash__(self):
        return self.data.__hash__()

class Blob(BaseNode):
    pass


class RemoteBranch(BaseNode):
    pass


class Tree(BaseNode):
    pass


class RemoteHead(BaseNode):
    pass


class Commit(BaseNode):
    pass


class RemoteServer(BaseNode):
    pass


class LocalBranch(BaseNode):
    pass


class AnnotatedTag(BaseNode):
    pass


class LocalHead(BaseNode):
    pass


class Tag(BaseNode):
    pass


# class Upstream(BaseNode):
#     pass



import datetime
import os
import networkx as nx
from .git_node import *
import git_graph.git_graph_class as gg
from copy import deepcopy

CURRENT_FOLDER = "."

ALL = "all"
COMMITS = "commits"
BRANCHES = "branches"

ALL_NODES = "dchatsglurb"
COMMIT_NODES = ALL_NODES.replace("b", "").replace("t", "")
BRANCH_NODES = COMMIT_NODES.replace("c", "")


def handle_specific_node_sets(nodes=ALL_NODES):
    if nodes == ALL:
        result = ALL_NODES
    elif nodes == COMMITS:
        result = COMMIT_NODES
    elif nodes == BRANCHES:
        result = BRANCH_NODES
    else:
        result = nodes
    return result


def filter_nodes(git_graph, nodes=ALL_NODES):
    node_set = set()
    if "b" in nodes:
        node_set.update(git_graph.blobs)
    if "t" in nodes:
        node_set.update(git_graph.trees)
    if "c" in nodes:
        node_set.update(git_graph.commits)
    if "l" in nodes:
        node_set.update(git_graph.local_branches)
    if "h" in nodes:
        node_set.update(git_graph.local_head[0])
    if "r" in nodes:
        node_set.update(git_graph.remote_branches)
    if "d" in nodes:
        node_set.update(git_graph.remote_heads)
    if "s" in nodes:
        node_set.update(git_graph.remote_servers)
    if "a" in nodes:
        node_set.update(git_graph.annotated_tags)
    if "g" in nodes:
        node_set.update(git_graph.tags)
    if "u" in nodes:
        pass
    return node_set


def resolve_child_type(node, parent, gg):
    if isinstance(parent, Tree):
        return Blob

    if isinstance(parent, Commit):
        if node in gg.commits:
            return Commit
        else:
            return Tree

    if isinstance(parent, Blob):
        return None

    if isinstance(parent, LocalBranch):
        return Commit

    if isinstance(parent, RemoteBranch):
        if node in gg.local_branches:
            return LocalBranch
        else:
            return Commit

    if isinstance(parent, RemoteServer):
        if node in gg.remote_heads:
            return RemoteHead
        else:
            return LocalBranch

    if isinstance(parent, Tag):
        if node in gg.commits:
            return Commit
        else:
            return AnnotatedTag

    if isinstance(parent, AnnotatedTag):
        return Commit

    if isinstance(parent, LocalHead):
        return LocalBranch

    if isinstance(parent, RemoteHead):
        return RemoteBranch

import pdb
class GitNX(nx.DiGraph):
    def __init__(self, git_path, nodes=ALL_NODES):
        super().__init__()
        self.git_path = git_path
        git_graph = gg.GitGraph(self.git_path).build_graph()
        self.git_graph=git_graph
        nodes = handle_specific_node_sets(nodes)
        node_set = filter_nodes(git_graph, nodes)
        if "b" in nodes:
            for _b in git_graph.blobs:
                self.add_node(Blob(_b))
        if "t" in nodes:
            for _t in git_graph.trees:
                t = Tree(_t)
                self.add_node(t)
                for e in git_graph.trees[_t]:
                    if e[0] in node_set:
                        self.add_edge(t, resolve_child_type(e, t, git_graph)(e[0]))
        if "c" in nodes:
            for _c in git_graph.commits:
                c = Commit(_c)
                self.add_node(c)
                for _e in git_graph.commits[_c]:
                    if _e in node_set:
                        self.add_edge(c, resolve_child_type(_e, c, git_graph)(_e))
        if "l" in nodes:
            for _l in git_graph.local_branches:
                l = LocalBranch(_l)
                self.add_node(l)
                e = git_graph.local_branches[_l]
                if e in node_set:
                    self.add_edge(l, resolve_child_type(e, l, git_graph)(e))
        if "h" in nodes:
            h = git_graph.local_head[0]
            h = LocalHead(h)
            self.add_node(h)
            e = git_graph.local_head[1]
            if e in node_set:
                self.add_edge(h, resolve_child_type(e, h, git_graph)(e))
        if "r" in nodes:
            for _r in git_graph.remote_branches:
                r = RemoteBranch(_r)
                self.add_node(r)
                e = git_graph.remote_branches[_r]
                if e in node_set:
                    self.add_edge(r, resolve_child_type(e, r, git_graph)(e))
        if "d" in nodes:
            for _d in git_graph.remote_heads:
                d = RemoteHead(_d)
                self.add_node(d)
                e = git_graph.remote_heads[_d]
                if e in node_set:
                    self.add_edge(d, resolve_child_type(e, d, git_graph)(e))
        if "s" in nodes:
            for _s in git_graph.remote_servers:
                s = RemoteServer(_s)
                self.add_node(s)
                for _e in git_graph.remote_servers[_s]:
                    if _e in node_set:
                        self.add_edge(s, resolve_child_type(_e, s, git_graph)(_e))
        if "a" in nodes:
            for _a in git_graph.annotated_tags:
                a = AnnotatedTag(_a)
                self.add_node(a)
                e = git_graph.annotated_tags[_a]
                if e in node_set:
                    self.add_edge(a, resolve_child_type(e, a, git_graph)(e))
        if "g" in nodes:
            for _g in git_graph.tags:
                g = Tag(_g)
                self.add_node(g)
                e = git_graph.tags[_g]
                if e in node_set:
                    self.add_edge(g, resolve_child_type(e, g, git_graph)(e))
        # if "u" in nodes:  # no color (only edges)
        #     for u in git_graph.upstreams:
        #         e = git_graph.upstreams[u]
        #         if e in node_set:
        #             self.add_edge(u, resolve_child_type(e, u, git_graph)(e))



import subprocess
import os

lbr = "refs/heads/"
rbr = "refs/remotes/"
tr = "refs/tags/"


def execute_git_command(path, command):
    if not os.path.isdir(os.path.join(path, ".git")):
        print("Not a git repository")
        return []
    bash_command = "git -C " + str(path) + " " + command
    try:
        output = subprocess.run(bash_command.split(), stdout=subprocess.PIPE)
    except subprocess.CalledProcessError:
        print("Not a git command")
        return []
    else:
        result = output.stdout.decode("utf-8").splitlines()
        return result


def read_git_file(path, sha1_file):
    return execute_git_command(path, "cat-file -p " + sha1_file)


def get_git_objects(path):
    objects = [
        line.split()
        for line in execute_git_command(
            path, "cat-file --batch-check --batch-all-objects"
        )
    ]
    blobs = [obj[0] for obj in objects if obj[1] == "blob"]
    trees = [obj[0] for obj in objects if obj[1] == "tree"]
    commits = [obj[0] for obj in objects if obj[1] == "commit"]
    annotated_tags = [obj[0] for obj in objects if obj[1] == "tag"]
    return blobs, trees, commits, annotated_tags


def get_git_references(path):
    references = [line.split() for line in execute_git_command(path, "for-each-ref")]
    local_branches = {
        ref[2][len(lbr) :]: ref[0] for ref in references if ref[2].startswith(lbr)
    }
    remote_branches = {
        ref[2][len(rbr) :]: ref[0]
        for ref in references
        if ref[2].startswith(rbr) and "/HEAD" not in ref[2]
    }
    tags = {ref[2][len(tr) :]: ref[0] for ref in references if ref[2].startswith(tr)}
    return local_branches, remote_branches, tags


def get_git_local_head(path):
    symbolic_ref = execute_git_command(path, "symbolic-ref HEAD -q")
    if symbolic_ref:
        return symbolic_ref[0][len(lbr) :]
    else:
        commit = execute_git_command(path, "rev-parse HEAD")
        return commit[0]


def get_git_remote_heads(path):
    lines = execute_git_command(path, "branch -rv --abbrev=0")
    lines_split = [line.split() for line in lines if "/HEAD " in line]
    remote_heads = {ls[0]: ls[2] for ls in lines_split}
    return remote_heads


def get_git_upstreams(path):
    lines = execute_git_command(path, "branch -vv --abbrev=0")
    lines_split = [line[2:].split() for line in lines if "[" in line]
    upstreams = {ls[2][1:-1]: ls[0] for ls in lines_split}
    return upstreams



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['git_networkx']

package_data = \
{'': ['*']}

install_requires = \
['flake8', 'networkx>=2.4,<3.0']

setup_kwargs = {
    'name': 'git-networkx',
    'version': '2.0.0',
    'description': 'Git graph to networkx',
    'long_description': '# Git-NetworkX\n\n### Port of https://github.com/hoduche/git-graph to NetworkX\n___\n\n## Install\n\n### PyPi\n`pip install git-graph`\n### Git\n`pip install git+https://github.com/CircArgs/git-networkx.git`\n\n## Demo\n\n```python\nfrom networkx.drawing.nx_pydot import write_dot\nimport git_networkx.git_networkx as gnx\n#everything\nAll = gnx.GitNX(\'my/repo/path/that/has/a/.git\')\nwrite_dot(All, "myrepo.dot")\n\n#Commits\nCommits = gnx.GitNX(\'my/repo/path/that/has/a/.git\', "c")\nwrite_dot(Commits, "mycommits.dot")\n\n```\n\n## Node Types\n| Node Type      | Letter |  Node Type      | Letter |\n| -------------- | :----: |  -------------- | :----: |\n| blob           | b      |  remote branch  | r      |\n| tree           | t      |  remote head    | d      |\n| commit         | c      |  remote server  | s      |\n| local branch   | l      |  annotated tag  | a      |\n| local head     | h      |  tag            | g      |\n| upstream link  | u      | \n\nBy default all nodes are added to the DiGraph.\n```python\n# you can get your commits, branches and the head of your local repo simply with lch\nG=gnx.GitNX(\'../git_networkx_test/\', "lch")\n\n```\n',
    'author': 'Henri-Olivier Duché',
    'author_email': 'hoduche@yahoo.fr',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/CircArgs/git-graph',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.6,<4.0',
}


setup(**setup_kwargs)
